#include <opencv2/core.hpp>

#include <iostream>
#include <fstream>
#include <vector>

int COMMON = 600;

struct Trafo {
	cv::Mat R;
	float s;
	cv::Point3f offset1;
	cv::Point3f offset2;
};

void readData(std::vector<cv::Point3f>& pc, std::string filename)
{
	std::ifstream f(filename);

	float x, y, z;
	int i = 0;
	while (!f.eof()) {
		if (f >> x >> y >> z)
			pc.push_back(cv::Point3f{ x, y, z });
		++i;
	}

	f.close();
}

cv::Point3f center(std::vector<cv::Point3f>& pc)
{
	cv::Point3f c{ 0.0, 0.0, 0.0 };
	for (int i = 0; i < COMMON; ++i)
	{
		c += pc[i];
	}

	float N = COMMON;
	c /= N;

	return c;
}

cv::Mat rotation(std::vector<cv::Point3f>& P, std::vector<cv::Point3f>& Q, Trafo& t)
{
	// Translation is eliminated using the Trafo.offset

	cv::Mat H = cv::Mat::zeros({ 3,3 }, CV_32F);

	// Q = [ q1^T 
	//       q2^T
	//       .
	//       .
	//       . ]
	// P = [ p1 p2 ... ]
	//
	// H = P*Q = p1*q1^T + p2*q2^T + .....
	//
	//  pi*qi^T = | p1 | * [ q1 q2 q3 ]
	//            | p2 |
	//            | p3 |
	//
	//

	for (int i = 0; i < COMMON; ++i) {
		float p1 = P[i].x - t.offset1.x;
		float p2 = P[i].y - t.offset1.y;
		float p3 = P[i].z - t.offset1.z;

		float q1 = Q[i].x - t.offset2.x;
		float q2 = Q[i].y - t.offset2.y;
		float q3 = Q[i].z - t.offset2.z;

		H.at<float>(0, 0) += p1 * q1;
		H.at<float>(0, 1) += p1 * q2;
		H.at<float>(0, 2) += p1 * q3;

		H.at<float>(1, 0) += p2 * q1;
		H.at<float>(1, 1) += p2 * q2;
		H.at<float>(1, 2) += p2 * q3;

		H.at<float>(2, 0) += p3 * q1;
		H.at<float>(2, 1) += p3 * q2;
		H.at<float>(2, 2) += p3 * q3;
	}

	cv::Mat U(3, 3, CV_32F), W(3, 3, CV_32F), Vt(3, 3, CV_32F);
	cv::SVD::compute(H, W, U, Vt, cv::SVD::Flags::FULL_UV);

	cv::Mat rot = Vt.t() * U.t();
	std::cout << "Determinant of Rotation: " << cv::determinant(rot) << std::endl;

	return Vt.t() * U.t();
}

float scale(std::vector<cv::Point3f>& P, std::vector<cv::Point3f>& Q, Trafo& t)
{
	//       SUM p^T * q
	// s =  -------------
	//       SUM p^T * p

	float numerator = 0.0;
	float denominator = 0.0;

	for (int i = 0; i < COMMON; ++i) {
		cv::Mat p(P[i] - t.offset1); // Eliminate translation
		p = t.R * p; // Eliminate rotation

		cv::Mat q(Q[i] - t.offset2); // Eliminate translation

		numerator += p.dot(q);
		denominator += p.dot(p);
	}

	return numerator / denominator;
}

int main()
{
	std::vector<cv::Point3f> P;
	std::vector<cv::Point3f> Q;


	readData(P, "data/sphere1.xyz");
	readData(Q, "data/sphere2.xyz");

	Trafo T;
	T.offset1 = center(P);
	T.offset2 = center(Q);

	T.R = rotation(P, Q, T);

	T.s = scale(P, Q, T);

	std::cout << T.offset1 << std::endl;
	std::cout << T.offset2 << std::endl;
	std::cout << T.R << std::endl;
	std::cout << T.s << std::endl;

	std::ofstream g("output.xyz");
	for (int i = 0; i < P.size(); ++i) {
		g   << Q[i].x << " "
			<< Q[i].y << " "
			<< Q[i].z << std::endl;

		cv::Mat p(3, 1, CV_32F);
		p.at<float>(0, 0) = P[i].x - T.offset1.x;
		p.at<float>(1, 0) = P[i].y - T.offset1.y;
		p.at<float>(2, 0) = P[i].z - T.offset1.z;

		p = (T.R * p) * T.s;

		p.at<float>(0, 0) += T.offset2.x;
		p.at<float>(1, 0) += T.offset2.y;
		p.at<float>(2, 0) += T.offset2.z;

		g << p.at<float>(0, 0) << " " << p.at<float>(1, 0) << " " << p.at<float>(2, 0) << std::endl;
	}
	g.close();

	return 0;
}


